/* The node_io level routines for the meme machine*/
/* Primitives defined in this file:

	write_to_scratch()
	compare_node_scratch()
	compare_node_node()
        add_to_chunk()
	take_from_chunk()	
	add_lhs_output()
	add_rhs_output()
	del_lhs_output()
	del_rhs_output()
*/

#include <stdio.h>
#include "datastru.h"

/**************************************************************************
 * write_to_scratch
 * This routine writes a node out to a scratch area. Used in compating
 * 2 nodes with each other.
 */

void write_to_scratch(net_type *net, int node)
{
   int lhs,rhs;
   
   if (node>BASE_NODE_SPACE)
   {
      lhs=net->data[node+LHS_IP];
      rhs=net->data[node+RHS_IP];
      write_to_scratch(net,lhs); write_to_scratch(net,rhs);
   }
   else
   {
     net->scratch_size++; net->scratch[net->scratch_size]=node;
     /* sanity check, remove this later for speed */
     if (net->scratch_size>=SCRATCH_SIZE) {printf("WOA!!! SC\n"); exit(0); }
   }
}

/***************************************************************************
 * int compare_node_scratch()
 * This recursive routine will compare a node with the net->scratch area
 */

int compare_node_scratch(net_type *net, int node)
{
   int lhs,rhs;
   int result;
   
   if (node>BASE_NODE_SPACE)
   {
      lhs=net->data[node+LHS_IP];
      rhs=net->data[node+RHS_IP];
      result=compare_node_scratch(net,lhs);
      if (result!=0) return(result);
      result=compare_node_scratch(net,rhs);
      return(result);
   }
   else /* base-node compare */
   {
  /* printf("compare '%c' with '%c'\n",
     node/8,net->scratch[net->scratch_posn]/8);*/
     if (net->scratch_posn>net->scratch_size) return(1);
     net->scratch_posn++;
     return(node-net->scratch[net->scratch_posn-1]);
   }
}


/***************************************************************************
 * int compare_node_node()
 * This routine compares one node with another. return is either 
 *  <0 (less than)  0 (exact) or >0 (greater than)
 * This routine is used for working out where in an output list a new
 * node should be inserted to keep order.
 * NOTE! We assume node1 has been written to the scratch area here.
 */

int compare_node_node(net_type *net, int node1, int node2)
{
  int first, result;
  
  /* printf("compare '"); print_node(net,node1); printf("' with '");
     print_node(net,node2); printf("'\n"); */
     
  /* for speed, compare first chars of 2 nodes only first*/
  first=net->data[node2]&FIRST_MASK;
  result=net->scratch[1]-first;
  /* printf("result=%i\n",result); */
  if (result!=0) return(result);
  
  /* now compare node2 with the contents of the scratch area.... */
  net->scratch_posn=1;
  result=compare_node_scratch(net,node2);
  if (result==0 && net->scratch_posn<=net->scratch_size) result=-1;
  return(result);
  
}


/****************************************************************************
 * int add_to_chunk(net,node,chunk,size)
 * This routine adds an item into a chunk (input or output) of a node.
 * size=old size of the chunk (net->num_node_ip[node] or net->num_node_op[node] )
 * Note that in this version, the order of the items in the chunk is kept.
 */

int add_to_chunk(net_type *net, int node, int chunk, int size)
{
  int new_chunk, loop;
  int bot,mid,top,insert_posn, result;
  
  /* printf("********* add '"); print_node(net,node); printf("' *****\n"); */
     
  /* special case - no initial outputs*/
  if (size==0)
  {
    new_chunk=get_chunk(net,1);
    net->data[new_chunk]=node;
    return(new_chunk);
  }
  
  new_chunk=chunk;
  switch(size)
  {
    case 1:
    case 2:
    case 4:
    case 8:
    case 16:
    case 32:
    case 64:
    case 128:
    case 256:
    case 512:
    case 1024: new_chunk=get_chunk(net,size*2); break;
    default: if (size%1024==0) new_chunk=get_chunk(net,size+1024);
  }
  
  
  /* now work out where in the chunk the new node lives... */
  bot=0; top=size-1; mid=0; result=1;
  /* printf("bot=%i, top=%i, size=%i\n",bot,top, size); */
  
  /* firstly, write the node to the scratch area... */
  net->scratch_size=0; write_to_scratch(net,node);
  while (bot<=top && result!=0)
  {
    mid=bot+(top-bot)/2;
    /* printf("mid=%i\n",mid); */
    result=compare_node_node(net,node,net->data[chunk+mid]);
    /* printf("result=%i\n",result); */
    if (result<0) bot=mid+1;
    if (result>0) top=mid-1;    
  }
  if (result==0) bot=mid;  /*keep exact matches next to each other*/
  insert_posn=bot;
 
  /* printf("insert_posn=%i\n",insert_posn); */
  if (new_chunk!=chunk)
    for (loop=0; loop<insert_posn; loop++)
         net->data[new_chunk+loop]=net->data[chunk+loop];
  for (loop=size-1; loop>=insert_posn; loop--)
  {
     net->data[new_chunk+loop+1]=net->data[chunk+loop];
  }
  net->data[new_chunk+insert_posn]=node;
  if (new_chunk!=chunk) free_chunk(net,chunk,size);  
  return(new_chunk);
}

/**********************************************************************
 * int take_from_chunk(net,node,chunk,size)
 * this routine removes a node from a chunk.
 * NOTE - this routine ALWAYS assumes that 'node' is present in 'chunk'.
 * If it is'nt, the whole program will exit.
 */
int take_from_chunk(net_type *net, int node, int chunk, int size)
{
  int new_chunk,loop;
  int skipped;
  int old_size=0;

  /* special case - only one item in chunk */
  if (size==1) {free_chunk(net,chunk,1); return(0);}
   
  skipped=FALSE;
  new_chunk=chunk;
  switch(size)
  {
    case 2:
    case 3:
    case 5:
    case 9:
    case 17:
    case 33:
    case 65:
    case 129:
    case 257:
    case 513:
    case 1025: {new_chunk=get_chunk(net,size-1); old_size=(size-1)*2; break;}
    default: if (size%1024==1)
                 {new_chunk=get_chunk(net,size-1); old_size=size-1+1024; }
  }
  for(loop=0; loop<size; loop++)
  {
    if (!skipped)  /*havent missed a item yet*/
    {
      if (net->data[chunk+loop]!=node)
	 net->data[new_chunk+loop]=net->data[chunk+loop]; 
      else
        skipped=TRUE;
    }
    else
      net->data[new_chunk+loop-1]=net->data[chunk+loop];   
  }
        
  if (new_chunk!=chunk) free_chunk(net,chunk,old_size);
  return(new_chunk);
}


/***************************************************************
 * add_lhs_output(net,node,output,side);
 * This routine adds a connection going out from 'node' into 
 * 'output'.
 * called by make_node()
 */

void add_lhs_output(net_type *net, int node, int output)
{
  int length;
  int chunk;

  chunk=net->data[node+LHS_OP];
  length=net->data[node+NUM_LHS_OP];
 
  net->data[node+LHS_OP]=add_to_chunk(net,output,chunk,length);
  net->data[node+NUM_LHS_OP]++;
}

/***************************************************************
 * add_rhs_output(net,node,output,side);
 * This routine adds a connection going out from 'node' into 
 * 'output'.
 * called by make_node()
 */

void add_rhs_output(net_type *net, int node, int output)
{
  int length;
  int chunk;
  
  chunk=net->data[node+RHS_OP];
  length=net->data[node+NUM_RHS_OP];
  
  net->data[node+RHS_OP]=add_to_chunk(net,output,chunk,length);  
  net->data[node+NUM_RHS_OP]++;
}

/***************************************************************
 * del_lhs_output(net,node,output);
 * This routine deletes a connection from 'node' to 
 * 'output'.
 * called by delete_node()
 */

void del_lhs_output(net_type *net, int node, int output)
{
  int length;
  int chunk;
  
  chunk=net->data[node+LHS_OP];
  length=net->data[node+NUM_LHS_OP];
  
  net->data[node+LHS_OP]=take_from_chunk(net,output,chunk,length);
  net->data[node+NUM_LHS_OP]--;
}

/***************************************************************
 * del_rhs_output(net,node,output);
 * This routine deletes a connection from 'node' to 
 * 'output'.
 * called by delete_node()
 */

void del_rhs_output(net_type *net, int node, int output)
{
  int length;
  int chunk;
  
  chunk=net->data[node+RHS_OP];
  length=net->data[node+NUM_RHS_OP];
  
  net->data[node+RHS_OP]=take_from_chunk(net,output,chunk,length);
  net->data[node+NUM_RHS_OP]--;
}
